import re
from datetime import timedelta
from urllib.parse import quote
from uuid import UUID

import django_rq
from botocore.exceptions import ClientError
from django.conf import settings
from django.core.management.base import BaseCommand
from django.db.models import Exists, Max, OuterRef
from django.utils import timezone
from rq.utils import as_text

from arkindex.documents.models import CorpusExport, CorpusExportState, Element
from arkindex.images.models import Image, ImageServer
from arkindex.ponos.models import Artifact, Task
from arkindex.process.models import (
    CorpusWorkerVersion,
    DataFile,
    Process,
    Worker,
    WorkerActivity,
    WorkerRun,
)
from arkindex.project.aws import s3
from arkindex.project.rq_overrides import Job
from arkindex.training.models import Model, ModelVersion
from redis.exceptions import ConnectionError

# Ponos artifacts use the path: <task id>/<path>
REGEX_ARTIFACT = re.compile(r"^(?P<task_id>[0-9a-f\-]{36})/")
# Ponos logs use the path: <task id>.log
REGEX_LOG = re.compile(r"^(?P<task_id>[0-9a-f\-]{36})\.log$")
# Model version artifacts use the path: <modelversion uuid>.zst
REGEX_MODEL_VERSION = re.compile(r"^(?P<modelversion_id>[0-9a-f\-]{36})\.zst$")


class Command(BaseCommand):
    help = "Clean up old corpus exports, trashed DataFiles, expired processes and S3 buckets"

    def handle(self, *args, **options):
        # Cleaning up workers could free some artifacts, so clean them before artifacts
        self.cleanup_archived_workers()

        self.cleanup_artifacts()

        self.cleanup_expired_processes()

        self.cleanup_old_exports()

        self.cleanup_orphan_exports()

        self.cleanup_trashed_datafiles()

        self.cleanup_orphan_images()

        self.cleanup_local_images()

        self.cleanup_ponos_logs()

        self.cleanup_archived_models()

        self.cleanup_unlinked_model_versions()

        self.cleanup_rq_user_registries()

    def cleanup_artifacts(self):
        """
        Remove all Ponos artifacts that are not tied to a Process
        """
        self.stdout.write("Removing orphaned Ponos artifacts…")

        bucket = s3.Bucket(settings.PONOS_S3_ARTIFACTS_BUCKET)
        for obj in bucket.objects.all():

            # Parse process and task
            match = REGEX_ARTIFACT.match(obj.key)
            if match is None:
                self.stdout.write(self.style.WARNING(f"Unsupported artifact {obj.key}"))
                continue

            # Find matching task
            matched_ids = match.groupdict()
            try:
                Task.objects.only("id").get(id=matched_ids["task_id"])
            except Task.DoesNotExist:
                # When no task is found, delete the S3 file
                self.stdout.write(f"Removing artifact {obj.key}…")
                try:
                    obj.delete()
                except ClientError as e:
                    self.stdout.write(self.style.ERROR(str(e)))

        self.stdout.write(self.style.SUCCESS("Successfully cleaned up orphaned Ponos artifacts."))

    def cleanup_expired_processes(self):
        expired_processes = Process \
            .objects \
            .annotate(max_expiry=Max("tasks__expiry")) \
            .filter(max_expiry__lt=timezone.now())

        # A Ponos task can be linked to a Dataset when it provides the artifacts for a Dataset in a Complete state.
        # As Datasets are meant to be stable, we do not want to destroy these tasks and artifacts.
        tasks = Task.objects.filter(process__in=expired_processes, dataset__isnull=True)
        artifacts = Artifact.objects.filter(task__in=tasks)

        self.stdout.write(f"Removing {artifacts.count()} artifacts of expired processes from S3…")
        for artifact in artifacts.select_related("task").iterator():
            self.stdout.write(f"Removing artifact {artifact.s3_object.key}")
            try:
                artifact.s3_delete()
            except ClientError as e:
                self.stdout.write(self.style.ERROR(str(e)))

        self.stdout.write(f"Removing logs for {tasks.count()} tasks of expired processes from S3…")
        for task in tasks.iterator():
            self.stdout.write(f"Removing task log {task.logs.s3_object.key}")
            try:
                task.logs.s3_delete()
            except ClientError as e:
                self.stdout.write(self.style.ERROR(str(e)))

        self.stdout.write(f"Removing {artifacts.count()} artifacts of expired processes…")
        artifacts.delete()

        self.stdout.write(f"Removing {tasks.count()} tasks of expired processes…")
        tasks.delete()

        # Expired processes are not actually deleted themselves, just emptied,
        # because deleting them implies removing WorkerRuns and WorkerActivities.
        self.stdout.write(self.style.SUCCESS("Successfully cleaned up expired processes."))

    def cleanup_old_exports(self):
        # Corpus exports that are over 2 weeks old
        old_exports = CorpusExport.objects.filter(created__lt=timezone.now() - timedelta(weeks=2)).only("id")

        # Only completed exports will have files on S3
        old_exports_with_files = old_exports.filter(state=CorpusExportState.Done)

        self.stdout.write(f"Removing {old_exports_with_files.count()} old corpus exports from S3…")
        for export in old_exports_with_files.iterator():
            self.stdout.write(f"Removing export {export.id} from S3…")
            try:
                export.s3_delete()
            except ClientError as e:
                if e.response["Error"]["Code"] == "404":
                    self.stdout.write(f"Export {export.id} not found on S3, skipping")
                else:
                    self.stdout.write(self.style.ERROR(str(e)))

        self.stdout.write(f"Removing {old_exports.count()} old corpus exports…")
        old_exports.delete()
        self.stdout.write(self.style.SUCCESS("Successfully cleaned up old corpus exports."))

    def cleanup_orphan_exports(self):
        """
        Remove all corpus exports on S3 that are not tied to a CorpusExport instance
        """
        self.stdout.write("Removing orphaned corpus exports…")

        bucket = s3.Bucket(settings.AWS_EXPORT_BUCKET)
        for obj in bucket.objects.all():
            try:
                export_id = UUID(obj.key)
            except ValueError:
                self.stdout.write(self.style.WARNING(f"Unsupported corpus export {obj.key}"))
                continue

            if not CorpusExport.objects.filter(id=export_id).exists():
                self.stdout.write(f"Removing corpus export {obj.key}…")
                try:
                    obj.delete()
                except ClientError as e:
                    self.stdout.write(self.style.ERROR(str(e)))

        self.stdout.write(self.style.SUCCESS("Successfully cleaned up orphaned corpus exports."))

    def cleanup_trashed_datafiles(self):
        datafiles = DataFile.objects.filter(trashed=True)
        self.stdout.write(f"Deleting {datafiles.count()} DataFiles marked as trashed from S3 and the database…")

        for datafile in datafiles:
            self.stdout.write(f"Deleting DataFile {datafile.id}…")
            try:
                datafile.s3_delete()
            except ClientError as e:
                if e.response["Error"]["Code"] == "404":
                    self.stdout.write(f"DataFile {datafile.id} not found on S3, skipping")
                else:
                    self.stdout.write(self.style.ERROR(str(e)))
            except Exception as e:
                self.stdout.write(self.style.ERROR(str(e)))
            else:
                datafile.delete()

        self.stdout.write(self.style.SUCCESS("Successfully cleaned up DataFiles marked as trashed."))

    def cleanup_local_images(self):
        """
        Remove all images on the S3 bucket of the local IIIF server that are not in the database
        """
        self.stdout.write("Removing orphaned local images…")

        bucket = s3.Bucket(ImageServer.objects.local.s3_bucket)
        for obj in bucket.objects.all():
            try:
                # Use quote() to translate S3 keys into IIIF paths.
                # safe='' ensures that slashes are translated into %2F
                ImageServer.objects.local.images.get(path=quote(obj.key, safe=""))
            except Image.DoesNotExist:
                # When no image is found, delete the S3 file
                self.stdout.write(f"Removing image {obj.key}…")
                try:
                    obj.delete()
                except ClientError as e:
                    self.stdout.write(self.style.ERROR(str(e)))

        self.stdout.write(self.style.SUCCESS("Successfully cleaned up orphaned local images."))

    def cleanup_orphan_images(self):
        """
        Remove all images in the database that are over 2 weeks old and are not tied to any element
        """
        self.stdout.write("Removing orphan images…")
        # In order to optimize the request, and not perform a join that reads all of the images and
        # elements in the db (Image.objects.filter(elements__isnull=True)), the queryset gets filtered
        # with an EXISTS query which looks for Element objects with an image_id that is the pk of
        # an Image object in the 'parent' queryset (OuterRef). Since we are looking for Image objects
        # with no related elements, we filter on `~Exists` to negate that.
        old_orphan_images = Image.objects.filter(
            ~Exists(
                Element.objects.filter(
                    image_id=OuterRef("pk")
                )
            )
        ) \
            .filter(created__lt=timezone.now() - timedelta(weeks=2)) \
            .only("id")
        delete_count = old_orphan_images.count()
        if delete_count > 0:
            self.stdout.write(f"Removing {delete_count} orphan images…")
            try:
                # Using _raw_delete() to stop Django from looking for related elements before deleting
                # since we know there are no related elements.
                old_orphan_images._raw_delete(using="default")
            except ClientError as e:
                self.stdout.write(self.style.ERROR(str(e)))
        self.stdout.write(self.style.SUCCESS("Successfully cleaned up orphan images."))

    def cleanup_ponos_logs(self):
        """
        Remove ponos logs that are not tied to a task in the database
        """
        self.stdout.write("Removing orphaned Ponos logs…")

        bucket = s3.Bucket(settings.PONOS_S3_LOGS_BUCKET)
        for obj in bucket.objects.all():

            # Parse process, run and task ID
            match = REGEX_LOG.match(obj.key)
            if match is None:
                self.stdout.write(self.style.WARNING(f"Unsupported log {obj.key}"))
                continue

            # Find matching task
            matched_ids = match.groupdict()
            try:
                Task.objects.only("id").get(id=matched_ids["task_id"])
            except Task.DoesNotExist:
                # When no task is found, delete the S3 file
                self.stdout.write(f"Removing log {obj.key}…")
                try:
                    obj.delete()
                except ClientError as e:
                    self.stdout.write(self.style.ERROR(str(e)))

        self.stdout.write(self.style.SUCCESS("Successfully cleaned up orphaned Ponos logs."))

    def cleanup_archived_workers(self):
        """
        Remove Worker instances that have been archived for longer than the configured worker cleanup delay
        and that are not being used in any worker result.
        """
        self.stdout.write("Removing archived workers…")

        workers = Worker.objects.filter(archived__lte=timezone.now() - timedelta(days=settings.WORKER_CLEANUP_DELAY))
        skipped, deleted = 0, 0
        for worker in workers.iterator():
            # There are both foreign keys for worker versions and worker runs on worker results.
            # Some old results might only have a worker version ID, but when a worker run ID is set,
            # the worker version ID is deduced from it, so we only have to check on the version.
            if worker.versions.all().in_use():
                skipped += 1
                continue

            # Skip any workers whose WorkerConfigurations are in use.
            # This should never happen since we already filter on the WorkerVersions,
            # but that could lead to deleting worker results when we didn't want to.
            if WorkerRun.objects.filter(configuration__worker=worker).in_use():
                self.stdout.write(self.style.WARNING(
                    f"Worker {worker.name} ({worker.id}) does not have any worker versions used by worker results, "
                    "but some of its worker configurations are in use."
                ))
                continue

            self.stdout.write(f"Removing worker {worker.name} ({worker.id})")
            worker.delete()
            deleted += 1

        if skipped:
            self.stdout.write(f"Skipping {skipped} archived workers that have worker versions or configurations used in worker results.")
        self.stdout.write(self.style.SUCCESS(f"Successfully cleaned up {deleted} archived workers."))

    def cleanup_archived_models(self):
        """
        Remove Model instances that have been archived for longer than the configured model cleanup delay
        and that are not being used in any worker result.
        """
        self.stdout.write("Removing archived models…")
        models = Model.objects.filter(archived__lte=timezone.now() - timedelta(days=settings.MODEL_CLEANUP_DELAY))

        skipped, deleted = 0, 0
        for model in models.iterator():
            if WorkerRun.objects.filter(model_version__model=model).in_use():
                skipped += 1
                continue

            self.stdout.write(f"Removing model {model.name} ({model.id})")

            # Remove CorpusWorkerVersions and WorkerActivities first
            # Those normally use SET_NULL, but this can cause the unique constraints to complain
            # if there already are rows with a model version set to None.
            WorkerActivity.objects.filter(model_version__model=model).delete()
            CorpusWorkerVersion.objects.filter(model_version__model=model).delete()

            model.delete()

            deleted += 1

        if skipped:
            self.stdout.write(f"Skipping {skipped} archived models that have model versions used in worker results.")
        self.stdout.write(self.style.SUCCESS(f"Successfully cleaned up {deleted} archived models."))

    def cleanup_unlinked_model_versions(self):
        self.stdout.write("Removing orphaned model versions archives…")
        bucket = s3.Bucket(settings.AWS_TRAINING_BUCKET)
        for obj in bucket.objects.all():

            # Parse model version ID
            match = REGEX_MODEL_VERSION.match(obj.key)
            if match is None:
                self.stdout.write(self.style.WARNING(f"Unsupported model version archive {obj.key}"))
                continue

            # Find matching archive
            matched_ids = match.groupdict()
            try:
                ModelVersion.objects.only("id").get(id=matched_ids["modelversion_id"])
            except ModelVersion.DoesNotExist:
                # When no model version archive is found, delete the S3 file
                self.stdout.write(f"Removing model version archive {obj.key}…")
                try:
                    obj.delete()
                except ClientError as e:
                    self.stdout.write(self.style.ERROR(str(e)))

        self.stdout.write(self.style.SUCCESS("Successfully cleaned up orphaned model versions archives."))

    def cleanup_rq_user_registries(self):
        """
        To link RQ tasks to users and make them available in ListJobs, we have a "user registry" which stores a list
        of RQ jobs IDs under rq:registry:user:<user ID>. This registry cannot be cleaned up automatically by RQ workers
        as they are not easily overridable; the cleanup is instead done when ListJobs is called, which means cleaning up
        a very large number of tasks can cause the API to timeout and the whole cleanup to be cancelled.
        To at least reduce the impact of this issue, the cleanup command also performs this cleanup for all users.
        """
        self.stdout.write("Cleaning up deleted jobs from RQ user registries…")
        deleted_jobs, user_count = 0, 0

        connection = django_rq.get_connection()

        # Skip cleaning up when Redis appears to be down.
        # No actual connection is made when calling get_connection, so we check on the first command we run
        try:
            registry_keys = connection.keys(pattern="rq:registry:user:*")
        except ConnectionError as e:
            self.stdout.write(self.style.ERROR(str(e)))
            return

        for registry_key in registry_keys:
            job_ids = list(connection.zrange(registry_key, 0, -1))
            to_delete = set()

            if not job_ids:
                # Empty registry
                continue

            # To check whether or not each job ID exists, we'll use the EXISTS command.
            # It accepts multiple names as arguments and returns how many of those names do exist,
            # but we need to know which job ID do exist or not, so we issue one EXISTS command per job.
            # To make this much faster, we use a pipeline to run many EXISTS commands and return all the results at once.
            with connection.pipeline() as pipeline:
                for job_id in job_ids:
                    pipeline.exists(Job.key_for(as_text(job_id)))

                for job_id, job_exists in zip(job_ids, pipeline.execute()):
                    if job_exists:
                        continue
                    to_delete.add(job_id)

                    # ZREM can be called with multiple keys at once, but with enough keys we'll get a broken pipe error,
                    # so to make sure we can really clean up a huge amount of job IDs, we delete in chunks.
                    if len(to_delete) >= settings.REDIS_ZREM_CHUNK_SIZE:
                        deleted_jobs += connection.zrem(registry_key, *to_delete)
                        to_delete = set()

            if to_delete:
                deleted_jobs += connection.zrem(registry_key, *to_delete)

            user_count += 1

        self.stdout.write(self.style.SUCCESS(f"Successfully cleaned up {deleted_jobs} deleted jobs from {user_count} RQ user registries."))
